//=============================================================================
// Unreal - Emitter.								Coded by .:..:
// Ver 2:
// - Using range instead of 2 variables will make it easier for mappers to configure it
// - Memory saving way can be used to save memory with particles.
// - Added particle revolution
// - No replication used (exept on network emitter)
//=============================================================================
class Emitter expands Actor;

struct FRange
{
	var() float Min,Max;
};
struct VRange
{
	var() FRange X,Y,Z;
};
struct IRange
{
	var() int Min,Max;
};
Enum ERangeType
{
	RT_Box,
	RT_Sphere,
	RT_Cylinder
};
struct HugeVecRange
{
	var() VRange BoxRange;
	var() FRange SphereRange;
	var() bool bPosRelativeToRot,bEffectedByLocation;
	var() ERangeType RangeType;
};
var(EmSpawning) HugeVecRange SpawnOffset;
var(EmSpawning) IRange NumberSpawnsBeforeDisabled;
var(EmSpawning) int MaxParticles;
var(EmSpeed) HugeVecRange ParticleAccel,ParticleVelocity;
var(EmTexture) texture EmitterTextures[8];
var(EmTexture) ERenderStyle EmitterStyle;
var(EmFading) FRange FadeInTime,FadeOutBeginTime,MaxFadeInScale;
var(EmSpawning) float NumParticlesPerSec;
var(EmMovement) bool bFallingParticles;
var(EmCollision) bool bEmitterHitWalls,bEmitterHitActors;
var(EmMovement) bool bBounchOnGround;
var(EmMovement) float EmitterBounchness;
var(EmMesh) Mesh EmitterMesh;
var(EmMesh) bool bAlwaysFaceCamera,bFaceVelocity,bEmitterEnviroMap;
var(EmMesh) rotator SpawnedRotation,MovingRotation;
var(EmMesh) name EmitterAnim;
var(EmMesh) float AnimationSpeed;
var(EmGeneral) bool bEnabled,bDisableWhenSpawned,bMemorySavingMode;
var(EmGeneral) float DistanceDisableRange;
var(EmGeneral) FRange ParticleScale,StayDisabledTime,ParticleLifeTime;
var(EmLight) byte LBrightness,LHue,LSaturation,LRadius;
var(EmLight) ELightType LType;
var(EmLight) ELightEffect LEffect;
var(EmSound) sound EmitterSound;
var(EmSound) byte SRadius,SPitch,SVolume;
var(EmRevolution) VRange RevolutionsPerSec;
var(EmRevolution) bool bUseRevolution;
var(EmRevolution) HugeVecRange RevolutionInitRate,RevolutionRangeScale;
var(EmTrigger) bool bToggleDisabled;
var(EmTrigger) IRange SpawnParticles;
var byte NumTex;
var int NumSpawned,NumParticlesNow,ParticlesSpawnedByMe,NumParticlesAllowedSpawn;
var bool bAnimate,bCheckDed,bCanBeDisabled,bTimerNeedsToReset;
var Player MyPlayer;
var float NextTimerPop,TimerPopRate;

replication
{
	// Things the server should send to the client.
	unreliable if( Role==ROLE_Authority )
		bEnabled;
}

simulated event Trigger( Actor Other, Pawn EventInstigator )
{
	if( bToggleDisabled )
		bEnabled = !bEnabled;
	else NumParticlesAllowedSpawn = ConvertIRange(SpawnParticles);
}

simulated function PostBeginPlay()
{
	local int i;
	
	if( Level.NetMode==NM_DedicatedServer )
		Return;
	Role = ROLE_Authority;
	Mesh = EmitterMesh;
	if( EmitterMesh==None || EmitterAnim=='' || !HasAnim(EmitterAnim) )
		bAnimate = False;
	else bAnimate = True;
	Mesh = None;
	bCanBeDisabled = (NumberSpawnsBeforeDisabled.Min!=0 || NumberSpawnsBeforeDisabled.Max!=0);
	if( bCanBeDisabled )
		NumSpawned = ConvertIRange(NumberSpawnsBeforeDisabled);
	SetUpTimer();
	for( i=0; i<8; i++ )
	{
		if( EmitterTextures[i]==None )
		{
			NumTex = i;
			Return;
		}
	}
	NumTex = 8;
}
simulated function SetUpTimer()
{
	local float Ti;
	
	TimerPopRate = -1;
	if( NumParticlesPerSec==0 )
		Ti = ParticleLifeTime.Max/MaxParticles;
	else Ti = 1/NumParticlesPerSec;
	if( Ti<0.1 )
	{
		Enable('Tick');
		TimerPopRate = Ti;
		NextTimerPop = Level.TimeSeconds+Ti;
	}
	else Disable('Tick');
	SetTimer(Ti,True);
	bTimerNeedsToReset = False;
}
simulated function Timer()
{
	if( !bCheckDed )
	{
		bCheckDed = True;
		MyPlayer = Class'EmX'.Static.GetPlayerOwner(Self);
		if( MyPlayer==None )
		{
			Disable('Tick');
			SetTimer(0,False);
			Return;
		}
	}
	if( !bEnabled && NumParticlesAllowedSpawn==0 )
	{
		bTimerNeedsToReset = True;
		SetTimer(1,False);
		TimerPopRate = -1;
		Disable('Tick');
		Return;
	}
	else if( bTimerNeedsToReset )
		SetUpTimer();
	if( SpawnAParticle() && bCanBeDisabled )
		NumSpawned--;
	if( bCanBeDisabled && NumSpawned<=0 )
	{
		NumSpawned = ConvertIRange(NumberSpawnsBeforeDisabled);
		bTimerNeedsToReset = True;
		if( bDisableWhenSpawned )
			bEnabled = False;
		else
		{
			SetTimer(ConvertFRange(StayDisabledTime),False);
			TimerPopRate = -1;
			Disable('Tick');
		}
	}
	if( NumParticlesAllowedSpawn>0 )
		NumParticlesAllowedSpawn--;
}
simulated function Tick( float Delta )
{
	local int w;
	
	While( NextTimerPop<Level.TimeSeconds )
	{
		NextTimerPop = NextTimerPop+TimerPopRate;
		Timer();
		if( TimerPopRate==-1 )
			Return;
		w++;
		if( w>=25 ) Return;
	}
}
simulated function bool SpawnAParticle()
{
	local EmitterParticle MyPart;
	local vector S;
	local float Di;
	
	if( MyPlayer==None ) Return False;
	if( DistanceDisableRange>0 )
	{
		Di = VSize(Class'EmX'.Static.GetPlayerCamLoc(MyPlayer)-Location);
		if( Di>DistanceDisableRange )
		{
			NextTimerPop = NextTimerPop+1;
			Return False;
		}
	}
	S = Location+ConvertHVRange(SpawnOffset,Rotation,Location);
	if( bMemorySavingMode )
	{
		MyPart = PickNextDeadParticle();
		if( MyPart==None && ParticlesSpawnedByMe<MaxParticles )
		{
			if( EmitterMesh==None )
				MyPart = Spawn(class'EmitterParticle',,,S);
			else MyPart = Spawn(class'MeshEmitter',,,S,SpawnedRotation);
			if( MyPart!=None )
				ParticlesSpawnedByMe++;
		}
		else if( MyPart!=None && MyPart.Location!=S )
			MyPart.SetLocation(S);
	}
	else if( EmitterMesh==None )
		MyPart = Spawn(class'EmitterParticle',,,S);
	else MyPart = Spawn(class'MeshEmitter',,,S,SpawnedRotation);
	if( MyPart!=None )
	{
		MyPart.StartingPos = S;
		CountMe(MyPart);
		SetUpParticle(MyPart);
		Return True;
	}
	Return False;
}
simulated function SetUpParticle( EmitterParticle P )
{
	local EPhysics E;
	local vector O;
	
	O = P.Location-Location;
	P.MyPlayer = MyPlayer;
	P.Texture = EmitterTextures[Rand(NumTex)];
	P.bBounce = bBounchOnGround;
	if( bFallingParticles )
		E = PHYS_Falling;
	else E = PHYS_Projectile;
	P.bSelfPhysics = (E==PHYS_None);
	if( P.Physics!=E )
		P.SetPhysics(E );
	if( P.bCollideActors!=bEmitterHitActors )
		P.SetCollision(bEmitterHitActors);
	P.bCollideWorld = (bEmitterHitWalls || bMemorySavingMode);
	P.bHasWorldCollision = bEmitterHitWalls;
	P.DrawScale = ConvertFRange(ParticleScale);
	P.Velocity = ConvertHVRange(ParticleVelocity,Rotation,O);
	P.Acceleration = ConvertHVRange(ParticleAccel,Rotation,O);
	P.FadeIn = ConvertFRange(FadeInTime);
	P.FadeOut = ConvertFRange(FadeOutBeginTime);
	P.FadeMax = ConvertFRange(MaxFadeInScale);
	P.LifeTime = Level.TimeSeconds+ConvertFRange(ParticleLifeTime);
	P.TimeStamp = Level.TimeSeconds;
	P.BounchNess = EmitterBounchness;
	P.Style = EmitterStyle;
	P.LightBrightness = LBrightness;
	P.LightHue = LHue;
	P.LightSaturation = LSaturation;
	P.LightRadius = LRadius;
	P.LightType = LType;
	P.LightEffect = LEffect;
	P.AmbientSound = EmitterSound;
	P.SoundRadius = SRadius;
	P.SoundPitch = SPitch;
	P.SoundVolume = SVolume;
	if( P.bDeadParticle )
	{
		P.bDeadParticle = False;
		P.Enable('Tick');
		P.bHidden = False;
	}
	P.bHandelRevolution = bUseRevolution;
	if( bUseRevolution )
	{
		P.CurLoc = P.Location;
		P.OlLoc = P.Location;
		P.RevRange = ConvertHVRange(RevolutionRangeScale,Rotation,O);
		P.InitRevP = ConvertHVRange(RevolutionInitRate,Rotation,O);
		P.RevPerSec = ConvertVRange(RevolutionsPerSec);
	}
	if( EmitterMesh!=None )
	{
		P.Skin = P.Texture;
		P.Mesh = EmitterMesh;
		P.bRotToVel = bFaceVelocity;
		P.bFaceCamera = bAlwaysFaceCamera;
		P.BeginRot = SpawnedRotation;
		if( !bFaceVelocity && !bAlwaysFaceCamera )
		{
			P.bFixedRotationDir = True;
			P.RotationRate = MovingRotation;
		}
		if( bAnimate )
			P.LoopAnim(EmitterAnim,AnimationSpeed);
		P.bMeshEnviroMap = bEmitterEnviroMap;
	}
}
simulated function CountMe( EmitterParticle Other )
{
	Other.MyEmitter = Self;
	NumParticlesNow++;
}
simulated function FreeMe( EmitterParticle Other )
{
	if( Other.MyEmitter==Self )
	{
		NumParticlesNow--;
		Other.MyEmitter = None;
	}
}
simulated function EmitterParticle PickNextDeadParticle()
{
	if( EmitterMesh==None )
		Return Class'EmX'.Static.PickNextDeadParticle(Class'EmitterParticle',XLevel);
	else Return Class'EmX'.Static.PickNextDeadParticle(Class'MeshEmitter',XLevel);
}
Static final function float ConvertFRange( FRange F )
{
	if( F.Min==F.Max )
		Return F.Min;
	Return F.Min+(F.Max-F.Min)*FRand();
}
Static final function vector ConvertVRange( VRange F )
{
	local vector V;
	
	V.X = ConvertFRange(F.X);
	V.Y = ConvertFRange(F.Y);
	V.Z = ConvertFRange(F.Z);
	Return V;
}
Static final function int ConvertIRange( IRange F )
{
	if( F.Min==F.Max )
		Return F.Min;
	Return F.Min+Rand(F.Max-F.Min);
}
Static final function vector ConvertHVRange( HugeVecRange F, optional rotator R, optional vector L )
{
	local vector V;
	
	if( F.bEffectedByLocation && L==vect(0,0,0) )
		L = VRand();
	if( F.RangeType==RT_Box )
		V = ConvertVRange(F.BoxRange);
	else if( F.RangeType==RT_Sphere )
	{
		if( F.bEffectedByLocation )
			V = Normal(L)*ConvertFRange(F.SphereRange);
		else V = VRand()*ConvertFRange(F.SphereRange);
	}
	else
	{
		if( F.bEffectedByLocation )
		{
			V = L;
			V.Z = 0;
		}
		else
		{
			V.X = 0.5-FRand();
			V.Y = 0.5-FRand();
		}
		V = Normal(V)*ConvertFRange(F.SphereRange);
	}
	if( F.bPosRelativeToRot )
		V = (V >> R);
	Return V;
}

defaultproperties
{
				MaxParticles=64
				EmitterStyle=STY_Translucent
				MaxFadeInScale=(Min=1.000000,Max=1.000000)
				bBounchOnGround=True
				EmitterBounchness=1.000000
				AnimationSpeed=1.000000
				bEnabled=True
				bDisableWhenSpawned=True
				bMemorySavingMode=True
				ParticleScale=(Min=1.000000,Max=1.000000)
				ParticleLifeTime=(Min=4.000000,Max=4.000000)
				SRadius=32
				SPitch=64
				SVolume=255
				bToggleDisabled=True
				bHidden=True
				bNoDelete=True
				bStasis=True
				RemoteRole=ROLE_None
				bDirectional=True
				Style=STY_Translucent
				Texture=Texture'FFG.Icons.S_Emitter'
}
